"""
Module to compute the strength of the material using the strength package
"""
import os
from pathlib import Path
import pandas as pd

import numpy as np

from calm.chemistry import ChemistrySizeDistribution
from calm.microstructure import Microstructure 
from strength.yieldstrength import YieldStrength
from premod.parser_generic import parse_complete


# what is the starting point for computing the Yield strength?
#   - microstructure state output
#   - a csv file with only average values
#   - history file


# Get microstructure object
# ============================

# Microstructure object
def get_default_matrix():
    """Create a Microstructure object (calm) from a microstructure dict 
    """
    micro = Microstructure(
        grainsize = 100.E-06,
        delta = 100.E-06, 
        rho_i = 1.E+10,   
        MTaylor = 3.1,   
    )
    return micro


# ============================
# Define precipitates object
# ============================
"""phases = {'beta_pp': {'Al': 5/33, 'Si': 14/33, 'Mg': 14/33},
         'U2': {'Al': 1/3, 'Si': 1/3, 'Mg': 1/3},
         'beta_p': {'Si': 3.33/9.33, 'Mg': 6/9.33},
         'Q_p': {'Al': 6/21, 'Si': 7/21, 'Mg': 6/21, 'Cu': 2/21},
         'L': {'Al': 20/99, 'Si': 33/99, 'Mg': 44/99, 'Cu': 2/99}}"""

def get_default_precipitate_state():
    """
    Default values to test the functions
    """
    metadata = {
        "ndomains": "5",
        "nelements": "6",
        "nphases": "1",
        "elements": ["Al","Mg","Si","Mn","Fe","Cu"],
        "phases": ["MG5SI6_B_DP_MYHR2001"],
        "X0":["0.98155092236025010","0.64466160664951319E-2","0.83177178803558374E-2","0.25927553985026675E-2","0.10202590093591746E-2","0.71729285037052328E-4"],
        #"XSS":[0.99218917962176767,0.12194061391042116E-3,0.39642000370768973E-2,0.26208562319053144E-2,0.10313167930845924E-2,0.72506702255074987E-4],
        "Xd_MG5SI6_B_DP_MYHR2001":["0.0000000000000000","0.58999999999999997","0.40999999999999998","0.0000000000000000","0.0000000000000000","0.0000000000000000"],
    }
    df = pd.DataFrame()
    df["DomainID[-]"] = [1,2,3,4,5]
    df["rpart[m]"] = [0.15000000200271479E-8,0.25000000333785799E-8,0.35000000467300119E-8,0.45000000600814438E-8,0.55000000734328758E-8]
    df["aspect[-]"] = [1.0,1.0,1.0,1.0,1.0]
    df["Nd[1/m3]"] = np.array([0.1,1.0,10.0,1.0,0.01])*100.0E+19
    df["phaseID[-]"] = [1,1,1,1,1]
    return metadata,df

def get_premod_state(filepath):
    """
    Parse a premod state and return the metada and dataframe
    """  
    metadata, df = parse_complete(filepath)

   # get in addition the time and the case name
 
    return metadata, df


def define_chem_premod(metadata:dict,df):
    """
    Define Chemistry object (calm) using data from precipitates state generated by premod.
    Return a ChemistrySizeDistribution object (defined in Calm)
    """
    # hard-coded phase name

    # Create chemistry object
    chem = ChemistrySizeDistribution(
        elements = metadata['elements'],
        X0 = np.asarray(metadata["X0"], dtype=float),
        phases = ['Al_matrix'],
    )
    
    # test to work only with single phase for now
    if int(metadata["nphases"]) > 1:
        raise ValueError("Only single phase microstructure allowed for now in define_chem_premod")

    #assuming only beta''
    #print(metadata)
    # make metadata["phases"] a list if not
    if isinstance(metadata['phases'],str):
        metadata['phases'] = [metadata['phases']]

    for idx, phasename in enumerate(metadata['phases']):
        #print("metadata in defin_chem_premod: ","Xd_"+phasename.strip(),metadata["Xd_"+phasename.strip()])
        chem.add_phase(
            name = phasename.strip(),
            Xp = np.asarray(metadata["Xd_"+phasename.strip()], dtype=float),
            rd = df["rpart[m]"].to_list(),
            Nd = df["Nd[1/m3]"].to_list(),
            rmean = None,
            rstd = None,
            shape = df["aspect[-]"].to_list(),
            distribution = None,
            reduce_volfrac=False,
        )
    
    #print("In define_chem_premod, chem_dict: \n", chem.__dict__)

    #comment = input("wainting for input")
    
    return chem


def define_chem_csv(filepath):
    """
    Define Chemistry object (calm) from a formatted csv file.
    Return a list of ChemistrySizeDistribution object (defined in Calm)
    """
    df_csv = pd.read_csv(filepath,sep=";",comment="#")
    columns = df_csv.columns.to_list()

    """
    X0_ ..., phasenames, MG5SI6_B_DP_MYHR2001_rv_avg (m),MG5SI6_B_DP_MYHR2001_Nv (m-3), Xd_MG5SI6_B_DP_MYHR2001_Al...
    """
    metadata = {"ndomains": 1}
    # extract elements
    elements = get_elements_columns(columns=columns)
    metadata["elements"] = elements
    metadata["nelements"] = len(elements)

    # extract phases
    phases = get_phases_columns(columns=columns)
    metadata["phases"] = phases
    metadata["nphases"] = len(phases)

    list_chems = []

    for irow,row in df_csv.iterrows():
        # Create chemistry object
        chem = ChemistrySizeDistribution(
        elements = metadata['elements'],
        X0 = get_X0_csv(metadata,row),
        phases = ['Al_matrix'],
        )
        for idx, phasename in enumerate(metadata['phases']):
            chem.add_phase(
                name = phasename,
                Xp = get_Xd_csv(metadata,row,phasename),
                rd = df_csv[phasename+"_rv_avg (m)"],
                Nd = df_csv[phasename+"_Nv (m-3)"],
                rmean = None,
                rstd = None,
                shape = df_csv[phasename+"_Shape"],
                distribution = None,
                reduce_volfrac=False,
            )
        list_chems.append(chem)
    
    return list_chems


def get_X0_csv(metadata,row):
    """
    Get the current initial composition from that row in the csv file
    """
    X0 = []
    for element in metadata["elements"]:
        label = "X0_"+element+" (at_frac)"
        try:# label in row.columns:
            X0.append(row[label])
        except:
            raise ValueError("Label: "+label+ " not found in columns names")
    return X0


def get_Xd_csv(metadata,row,phasename):
    """
    Get the current composition of the phase from that row in the csv file
    """
    Xd = []
    for element in metadata["elements"]:
        label = phasename+"_Xd_"+element+" (at_frac)"
        try:# label in row.columns():
            Xd.append(row[label])
        except:
            raise ValueError("Label: "+label+ " not found in columns names")
    return Xd


def get_elements_columns(columns:list):
    """
    Extract from the list the list of elements
    Return: list of elements
    """
    elements = []

    for label in columns:
        if label.startswith("X0_"):
            elements.append(label[3:].split(" ")[0])
    return elements


def get_phases_columns(columns:list):
    """
    Extract the name of the phases from the columns name
    Return list of phases name
    """
    phases = []
    
    for label in columns:
        if label.endswith("_Nv (m-3)"):
            phases.append(label.split("_Nv (m-3)")[0])
    return phases


# Calculate YieldStrength
# ============================
def calculate_YS_premod(metadata:dict,df,micro=None,
                 model='Myhr2015', terms=True):
    """
    Function computing the yield strength of a microstructure for a given model.
    Function expecting instances of entities as dict as input parameters.
    Returns total yield and tau for each contribution term.

    """
    if micro is None:
        micro = get_default_matrix()
    
    chem = define_chem_premod(metadata,df)

    s_tot, tau_terms = calculate_YS(chem,micro,model=model, terms=terms)

    s_terms = {}
    for term, value in tau_terms.items():
        s_terms[term] = value * micro.MTaylor
    return s_tot, s_terms


def calculate_YS_csv(filepath,micro=None,
                 model='Myhr2015', terms=True):
    """
    Function computing the yield strength of a microstructure for a given model.
    Function expecting instances of entities as dict as input parameters.
    Returns total yield and tau for each contribution term.

    """
    if micro is None:
        micro = get_default_matrix()
    
    list_chems = define_chem_csv(filepath)

    list_Ys = []
    for chem in list_chems:
        s_tot, tau_terms = calculate_YS(chem,micro,model=model, terms=terms)
        s_terms = {}
        for term, value in tau_terms.items():
            s_terms[term] = value * micro.MTaylor
        list_Ys.append([s_tot, s_terms])
    return list_Ys


# Calculate YieldStrength
# ============================
def calculate_YS(chem,micro,model='Myhr2015', terms=True):
    """
    Function computing the yield strength of a microstructure for a given model.
    Function expecting instances of entities as dict as input parameters.
    Returns total yield and tau for each contribution term.
    """
    yld = YieldStrength(chem, micro)
    yld.models['ppt'] = model
    strength = yld.yieldstrength(terms = terms)
    s_tot, tau_terms = strength
    return s_tot, tau_terms


def YS_to_HV_AA6xxx(yield_stress):
    """
    Calculation of the Vickers and  hardness based on the yield stress

    The coefficient for Vickers hardness are coming from fitting of 
    some measurements in Ampere
    """
    hardness = {
        "Vickers": 27 + 0.24 * yield_stress,
        "Rockwell B": 50 + 0.32 * yield_stress,
    }

    return hardness


if __name__ == "__main__":

    print("local testing")
    #test_getDistribution()
    #test_defineChem()

    #test_calculate_YS()



# -*- coding: utf-8 -*-
"""
    Plot the results of a Premod simulation
"""

import numpy as np
import matplotlib.pyplot as plt

from premod.io import Results


class PlotResults(object):
    
    # curve type dictionary: key=curve name/type, value=curve label
    curves = {'standard_f': 'Volume fraction vs Time',
              'standard_r': 'Mean radius vs Time',
              'chemistry': 'Composition vs Time'}
    
    def __init__(self, result=None):
        self.result = Results() if result is None else result
        
    def plot(self, figtyp, fig_kw=dict()):
        subplots = {1: 111, 2: 121, 3: 221, 4: 221}
        types = figtyp.split(',')
        n = len(types)
        if n in subplots:
            gca = fig_kw.pop('gca', False) # use the current axes
            subplot = 0 if gca else subplots[n]
            for typ in types:
                func = getattr(self, 'plot_' + typ, None)
                if func is not None:
                    func(subplot=subplot, **fig_kw)
                else:
                    err = 'Figure type unknown: "{}"'.format(figtyp)
                    raise RuntimeError(err)
                if subplot > 0:
                    subplot += 1
            
    def _plot_standard(self, typ, ax1, ax2,both_axes=True):
        """
        Function to generate a 'standard' plot meaning the time evolution
        of some variables either fv (-);rv_avg (m);Nv (m-3) (old version of Premod)
        in new version MG5SI6_B_DP_fv (-);MG5SI6_B_DP_rv_avg (m); ;MG5SI6_B_DP_Nv (m-3)
        """
        if 'Time' not in self.result.history:
            return
        x = self.result.history['Time']
        if self.result.history.IO_version == 1:
            yname = 'fv' if typ == 'f' else 'rv_avg'
            ylabel = r'$f_v \,\, (-)$' if typ == 'f' else r'$\overline{r}_{v} \,\, (m)$'
            leg_label = r'$f_v$' if typ == 'f' else r'$\overline{r}_{v}$'
            y1 = self.result.history[yname]
            y2 = self.result.history['Nv']
            leg_label2 = r'$N_v$'
        elif self.result.history.IO_version == 2:
            # hard-coded phase
            phase = "MG5SI6_B_DP"
            yname = 'fv' if typ == 'f' else 'rv_avg'
            yname = phase + "_" + yname
            ylabel = r'$f_v \,\, (-)$' if typ == 'f' else r'$\overline{r}_{v} \,\, (m)$'
            leg_label = r'$f_v$' if typ == 'f' else r'$\overline{r}_{v}$'
            leg_label = leg_label + " " + phase
            y1 = self.result.history[yname]
            yname2 = phase+"_Nv"
            y2 = self.result.history[yname2]
            leg_label2 = r'$N_v$' + " " + phase
        else:
            raise ValueError("unrecognized version: ",self.result.history.IO_version)
        
        ax1.plot(x, y1, label=leg_label, color='red')
        ax1.grid(True)
        ax1.set_xscale('log')
        ax1.set_xlabel(r'$Time \,\, (s)$', fontsize=14)
        ax1.set_ylabel(ylabel, fontsize=14)
        lines, labels = ax1.get_legend_handles_labels()
        if both_axes:
            ax2.plot(x, y2, label=leg_label2, color='blue')
            ax2.set_ylabel(r'$N_{v} \,\, (m^{-3})$', fontsize=14)
            lines_2, labels_2 = ax2.get_legend_handles_labels()
        
            lines += lines_2
            labels += labels_2
        
        ax1.legend(lines, labels, loc='best')
        
    def _gca(self, twinx, subplot, **fig_kw):
        """ Create or get the axes to do the plot """
        if subplot:
            plt.figure(**fig_kw)
            ax1 = plt.subplot(subplot)
        else:
            ax1 = plt.gca()
        if twinx:
            return ax1, ax1.twinx()
        else:
            return ax1
        
    def plot_standard_f(self, subplot=111,both_axes=True, **fig_kw):
        """ Function to plot in a standard way data from the ppt history file
            generated by Premod
        """
        ax1, ax2 = self._gca(True, subplot, **fig_kw)
        self._plot_standard('f', ax1, ax2,both_axes=both_axes)
        
    def plot_standard_r(self, subplot=111,both_axes=True, **fig_kw):
        """ Function to plot in a standard way data from the ppt history file
            generated by Premod
        """
        ax1, ax2 = self._gca(True, subplot, **fig_kw)
        self._plot_standard('r', ax1, ax2,both_axes=both_axes)
        
    def plot_chemistry(self, subplot=111, **fig_kw):
        """ Plot composition as function of Time from the ppt history file
            generated by Premod
        """
        self._gca(False, subplot, **fig_kw)
        if 'Time' in self.result.history:
            x = self.result.history['Time']
            if self.result.history.IO_version == 1:
                label_chem = "X_"
            elif self.result.history.IO_version == 2:
                label_chem = "XSS_"
            else:
                raise ValueError("unrecognized version: ",self.result.history.IO_version)
            for key, val in self.result.history.items():
                if label_chem in key:
                    label = key.split(label_chem)[1]
                    if label != 'Al':
                        plt.plot(x, val, label=r'$'+label+'$')
            plt.xscale('log')
            plt.xlabel(r'$Time \,\, (s)$', fontsize=14)
            plt.ylabel(r'$Composition \,\, (at)$', fontsize=14)
            plt.legend(loc='best')
            plt.grid(True)

def plot_sizedistribution(df, subplot=111, **fig_kw):
    """ Plot size distribution from the microstructure file
        generated by Premod
    """
    ax1 = plt.gca()
    if 'rpart[m]' in df.columns:
        x = df['rpart[m]']
        y = df['Nd[1/m3]']
        plt.plot(x, y, label='Nd[1/m3]')
        #plt.xscale('log')
        plt.xlabel('Particle radius (m)', fontsize=14)
        plt.ylabel('Particle density (1/m3)', fontsize=14)
        plt.legend(loc='best')
        plt.grid(True)

    
        


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import sys
import json
import dlite
from pathlib import Path

# standard plot
def standardplotVf():
    """
    """
    for element in fracs:
        plt.plot(times, df[element], label = element)
    plt.legend()
    plt.xlabel('Time (s)')
    plt.ylabel('Volume fraction')
    plt.show()
    return

# pie chart

def piechartVf():
    """
    """
    fig, ax = plt.subplots(ncols=len(times), figsize = (12,3))
    for i, row in df.iterrows():
        ax[i].pie(np.array([row['X_Mg (at_frac)'], row['X_Si (at_frac)'], row['X_Mn (at_frac)'], row['X_Fe (at_frac)']]), 
                labels = ['Mg', 'Si', 'Mn', 'Fe'],
                startangle = 90)
        ax[i].set_title(f"t = {int(row['Time (s)'])}")
    plt.show()
    return

def rv_Nd():
    """
    """
    plt.plot(df['rv_avg (m)'], df['Nv (m-3)'])
    plt.xlabel('rv')
    plt.ylabel('Nv')
    plt.show()
    return

def formatting_hist(): #a modifier
    """
    """
    # creer entite
    local_dir = Path(__file__).parent.resolve()
    os.chdir(local_dir)
    entity_path = "./entities/"   \
                  "entity_particlesizedistribution-0.1.json"
    #print(entity_path)
    dlite.storage_path.append(f'{entity_path}')
    alloyEntity = Instance.from_url(entity_path)

    # charger donnees read_csv
    df = pd.read_csv('fichier', sep=';', comment='#')

    # ouvrir le ficher txt pour avoir la list des lignes qui commence pqr diese pour recuperer les compo de la phase
    # convertir en dict 
    df.to_dict()

    # extract data
    # Time (s);Temp (K);fv (-);rv_avg (m);Nv (m-3);X_Al (at_frac);X_Mg (at_frac);X_Si (at_frac);X_Mn (at_frac);X_Fe (at_frac);X_Cu (at_frac)
    # list element;
    # nom des phases a ajouter dans la partie fortran, compo fixee
                                                                                                                                                                          
    results = {}
    results["dimensions"] = []
    results["dimensions"].append({"nelements" : nelements})
    results["nphases"].append({"nphases" : nphases})
    results["ntimestep"].append({"ntimestep" : ntimestep})

    results["properties"] = []
    results["properties"].append({"elements" : elements})
    results["properties"].append({"phases" : phases})
    results["properties"].append({"X0" : X0})
    results["properties"].append({"Xp" : Xp})
    results["properties"].append({"volfrac" : volfrac})
    results["properties"].append({"rmean" : rmean})
    results["properties"].append({"atvol" : atvol})
    
    with open(example + 'results', "w") as fp:
        json.dump(results, fp, indent=4) 
    return results


def formatting_chem():
    """
    """
    results = {}
    results["dimensions"] = []
    results["dimensions"].append({"nelements" : nelements})
    results["nphases"].append({"nphases" : nphases})

    results["properties"] = []
    results["properties"].append({"elements" : elements})
    results["properties"].append({"phases" : phases})
    results["properties"].append({"X0" : X0})
    results["properties"].append({"Xp" : Xp})
    results["properties"].append({"volfrac" : volfrac})
    results["properties"].append({"rmean" : rmean})
    results["properties"].append({"atvol" : atvol})
    
    with open(example + 'results', "w") as fp:
        json.dump(results, fp, indent=4) 
    return results


def test_read_results():
    dirname = os.path.dirname(__file__)
    example = 'TestMyhr2001'

    df = pd.read_table(dirname+'\\'+example+'\\'+example+'_summary.txt', sep=';', encoding='utf8')
    #print(df.head())
    #print(df.keys())
    times = list(df['Time (s)'])
    fracs = ['X_Mg (at_frac)', 'X_Si (at_frac)', 'X_Mn (at_frac)', 'X_Fe (at_frac)']



#-------------------------------------------------------------------------------
if __name__ == "__main__":
    
    standardplotVf()
    rv_Nd()


# -*- coding: utf-8 -*-
"""
    Execute the Premod solver
"""

import os
import numpy as np
import subprocess

from premod.io import PremodInputs, Results 
from premod.plot import PlotResults
#from pario.chemistry import Chemistry


class Progress(object):
    
    """ Simple object to track the status of a simulation """
    
    def __init__(self):
        self.status = ''
        self.message = ''
        
    def failed(self, message):
        self.status = 'failed'
        self.message = message
        
    def success(self, message=''):
        self.status = 'success'
        self.message = message
    
    def __call__(self, percent, message):
        pass


class PremodSolver(object):
    
    """
        
        Class to prepare the inputs, read the outputs and run a premod 
        simulation.
        
        Open a template, change parameters, save input in a different location,
        run the simulation, plot the results:
        
        >>> premod = PremodSolver(filename='template.txt', exe='premod.exe')
        >>> premod.set_weight_fraction('Mg', 0.82)
        >>> premod.run('test.txt')
        >>> premod.plot(figtyp='particle_evolution')
            
    """
    
    name = 'premod'
    
    def __init__(self, inp=None, filename=None, exe=''):
        self.exe = exe
        self.inp = PremodInputs()
        self.chemistry = None
        self.result = Results()
        self.fig = PlotResults(self.result)
        if inp is not None and filename is not None:
            raise ValueError("initialization of PremodSolver with both filename and inp")
        if inp is not None:
            self.inp = inp
        if filename is not None:
            self.read(filename)
        
    def _update_alloy(self):
        """ Update the alloy input file from the chemistry object """
        unit = self.inp.alloy['UNIT']
        el = self.chemistry.symbols
        if unit.find('w') == 0:
            frac = self.chemistry.weight_fractions
        elif unit.find('a') == 0:
            frac = self.chemistry.atom_fractions
        if unit.find('%') > 0:
            frac *= 100.0
        self.inp.alloy.set_table(cols=[el, frac])
            
    def _create_chemistry(self, elements, fractions, unit):
        """ Create a chemistry object for the alloy composition """
        if 'Al' in elements:
            elements.remove('Al')
        elements.insert(0, 'Al')
        frac = np.array(fractions)
        if unit.find('%') > 0:
            frac *= 0.01
        if unit.find('w') == 0:
            self.chemistry = Chemistry(elements, weight_fractions=frac)
        elif unit.find('a') == 0:
            self.chemistry = Chemistry(elements, atom_fractions=frac)
            
    def set_alloy(self, name, elements, fractions, unit='wt%'):
        """ Set the composition of the alloy """
        self.inp.alloy['ALLOYNAME'] = name
        self.inp.alloy['UNIT'] = unit
        self._create_chemistry(elements, fractions, unit)
    
    def set_weight_fraction(self, element, fraction):
        """ Set the weight fraction of the given element """
        self.chemistry.set_weight_fraction(element, fraction)
        self._update_alloy()
    
    def set_atom_fraction(self, element, fraction):
        """ Set the atom fraction of the given element """
        self.chemistry.set_atom_fraction(element, fraction)
        self._update_alloy()
        
    def set_fraction(self, element, fraction, unit):
        """ Set the element fraction with the specified unit """
        if unit.find('%') > 0:
            fraction *= 0.01
        if unit.find('w') == 0:
            self.chemistry.set_weight_fraction(element, fraction)
        elif unit.find('a') == 0:
            self.chemistry.set_atom_fraction(element, fraction)
        self._update_alloy()
        
    def fractions(self, unit):
        """ Return the fractions in wt, wt% or at_frac """
        if unit.find('w') == 0:
            f = self.chemistry.weight_fractions
        elif unit.find('a') == 0:
            f = self.chemistry.atom_fractions
        if unit.find('%') > 0:
            f *= 100.0
        return f
    
    def balance(self, unit):
        """ Return the balance (Al) in wt, wt% or at_frac """
        fractions = self.fractions(unit)
        return fractions[0]
    
    def composition(self, unit):
        """ Return a dict with key: symbol, value: fraction """
        return dict(zip(self.chemistry.symbols, self.fractions(unit)))
        
    def set_process(self, time, temperature):
        """ Set the temperature history / thermal treatment """
        self.inp.process.set_table(cols=[time, temperature])
        
    def set_output(self, times=None, period=None):
        if isinstance(times, (list, tuple, np.ndarray)):
            self.inp['OUTPUT_TIMES'] = [str(x) for x in times]
        if period is not None:
            try:
                p = float(period)
            except:
                p = None
            if p is not None:
                self.inp['OUTPUT_PERIOD'] = p

    def set_instance_IO_version(self,instance_IO_version=1):
        """ set the value for the instance_IO_version parameter"""
        self.inp['INSTANCE_IO_VERSION'] = instance_IO_version
    
    def set_phase(self, name, ppt):
        """ Set the table for the PPTSIM input file
        
            name: str
                Name of one phase model defined in the PHASES file
                
            ppt: str
                Name of one precipitate model defined in the PPTLIB file
        
        """
        col = ['', '']
        if isinstance(name, int):
            col[0] = self.inp.phases.entry[name]['NAME']
        elif isinstance(name, str):            
            for entry in self.inp.phases.entry:
                if entry['NAME'] == name:
                    col[0] = name
                    break

        if isinstance(ppt, int):
            col[1] = self.inp.pptlib.entry[name]['NAME']
        elif isinstance(name, str):            
            for entry in self.inp.pptlib.entry:
                if entry['NAME'] == ppt:
                    col[1] = ppt
                    break
        
        if all([len(x) for x in col]): 
            self.inp.pptsim.set_table(cols=[col])
        else:
            if not col[0]:
                raise RuntimeError('Phase not found: ' + name)
            if not col[1]:
                raise RuntimeError('Precipitate not found: ' + ppt)

    def read(self, filename):
        """ Read the input files """
        self.inp.read(filename)
        elements = self.inp.alloy.column(0)
        fractions = self.inp.alloy.column(1, 'float')
        self._create_chemistry(elements, fractions, self.inp.alloy['UNIT']) 
        self.result.read(filename, self.inp)

    def write(self, filename):
        """ Write the input files """
        self.inp.write(filename)
        
    def run(self, filename, progress=None):
        """ Write the input files, run the premod solver and read the outputs
            
            filename: str
                Input path and filename
                
            progress: Progress object
                Track the simulation status and show progress
        """
        return self._run(filename, add_input=True, progress=progress)
        
    def check_log(self, line, progress):
        """ Check the output line of the solver and change the status of the
            progress object
        """
        if line.find('fatal') == 0:
            error = line.replace('The program will stop', '')
            error = error.replace('fatal', '').strip()
            progress.failed(error)
            
        elif line.find('error') == 0:
            error = line.replace('The program will stop', '')
            error = error.replace('error', '').strip()
            progress.failed(error)
            
        elif line.find('Written: Output') >= 0:
            parts = line.split('=')
            tim = float(parts[1])
            if tim not in progress.times:
                progress.times.add(tim)                        
                done = 100.0 * len(progress.times) / len(progress.out_times)
                progress(done, line.replace('info', '').strip())
                    
        elif line.find('Program ended normally') >= 0:
            progress.success()
    
    def _run(self, filename, add_input=True, progress=None):
        self.inp.write(filename)
        self.result.clear()
        # run process    
        basename = os.path.basename(filename)
        casedir = os.path.dirname(filename)
        if casedir == '':
            casedir = './'
        args = [self.exe, basename] if add_input else [self.exe]
        
        if progress is None:
            progress = Progress()
              
        with subprocess.Popen(args, cwd=casedir,
                              stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              universal_newlines=True) as proc:

            progress.message = ''
            progress.status = ''
            progress.times = set()
            progress.out_times = [float(x) for x in self.inp['OUTPUT_TIMES']]
            while True:
                self.check_log(proc.stdout.readline(), progress)
                if progress.status == 'failed':
                    break
                if proc.poll() is not None:
                    break            
            proc.terminate()
    
        if progress.status == 'success':
            self.result.read(filename, self.inp)
        
        return progress.status, progress.message
        
    def plot(self, figtyp='', fig_kw=dict()):
        """ Plot a result
            
            figtyp: str
                Type of the figure

            fig_kw: str
                Figure parameters, see matplotlib.pyplot.figure
        """
        self.fig.plot(figtyp, fig_kw)


class Chemistry(object):

    def __init__(self,elements,
                weight_fractions=None,
                atom_fractions=None):
        self.elements = elements
        self.weight_fractions = weight_fractions
        self.atom_fractions = atom_fractions


""" Parser of the csv file with header"""
import pandas as pd


default_config = {
    "sep": ",",
    "key_sep": ":",
    "comment": "#"
}


def parse_complete(filepath,config=default_config):
    """
    Function to perform the complete parsing of the 
    file and return the metadata and datafranme
    """

    metadata = parse_metadata(filepath=filepath,config=config)

    df = parse_table(filepath=filepath, columns=metadata["columns"]
                    ,config=default_config)

    return metadata,df


def get_header(filepath,config=default_config):
    """ get the header lines"""
    lines = []
    with open(filepath) as f:
        line = f.readline()
        while line.startswith(config["comment"]):
            lines.append(line.lstrip(config["comment"]).strip())
            line = f.readline()
    return lines


def parse_metadata(filepath,config=default_config):
    """ parse the metadata in the header"""
    metadata = {}
    lines = get_header(filepath,config)
    for line in lines[:-1]:
        if config["key_sep"] in line:
            ipos = line.find(config["key_sep"])
            if ipos != -1:
                key = line.split(config["key_sep"])[0].strip()
                values = line[ipos+1:].split(config["sep"])
                if len(values) > 1:
                    metadata[key] = values
                else:
                    metadata[key] = values[0].strip()

    metadata["columns"] = lines[-1].split(config["sep"])

    return metadata


def parse_table(filepath,columns,config=default_config):
    """ parse the table of simulations"""
    df = pd.read_csv(filepath,sep=config["sep"],comment=config["comment"]
                    ,names=columns)

    return df



# -*- coding: utf-8 -*-
"""
    Handling of input files and output files of Premod
"""

import os, sys
from glob import glob
from collections import OrderedDict
from inspect import isfunction, ismethod

import numpy as np

dirname = os.path.dirname(__file__)
sys.path.append(os.path.join(dirname, '..', '..','..','calm'))
from calm.chemistry import ChemistrySizeDistribution
from premod.parser_generic import parse_complete
from premod.strength import define_chem_premod


class InputFile(OrderedDict):

    def __init__(self, name):
        OrderedDict.__init__(self)
        self.name = name
        self.comment = '!'

    def read(self, filename):
        """ Read a input file """
        with open(filename) as fil:
            self.read_string(fil.read())

    def read_string(self, value):
        """ Read a input string """
        pass

    def to_string(self):
        """ Convert a InputFile to a string """
        return ''

    def write(self, filename):
        """ Write a input file """
        with open(filename, 'w') as fil:
            fil.write(self.to_string())


class ConfigFile(InputFile):

    def __init__(self, name, key='', keep_comment=False):
        InputFile.__init__(self, name)
        self.keep_comment = keep_comment        
        self.key = key
        self.comments = []
        self.sequence = []

    def read_string(self, value):
        self.clear()
        self.comments.clear()
        self.sequence.clear()
        lines = value.split('\n')
        i = 0
        while i < len(lines):
            line = lines[i]
            if line.find(self.comment) == 0:
                self.sequence.append(len(self.comments))
                self.comments.append(line)
            else:
                delimiter = '='
                kv = [x.strip() for x in line.split('=')]
                if len(kv) < 2:
                    kv = [x.strip() for x in line.split()]
                    delimiter = ' '
                if len(kv) == 2:
                    key, val = kv
                    if key == self.key:
                        i += 1
                        val = [x for x in lines[i].split()]
                    self[key] = val
                    self.sequence.append([key, delimiter])
            i += 1

    def to_string(self):
        """ Format the ConfigFile to a string """
        lines = []
        for key in self.sequence:
            if isinstance(key, int):
                if self.keep_comment:
                    lines.append(self.comments[key])
            else:
                k = key[0]
                delimiter = key[1]
                v = self[k]
                if k == self.key:
                    lines.append('{}{}{}'.format(k, delimiter, len(v)))
                    lines.append(' '.join([str(x) for x in v]))
                else:
                    lines.append('{}{}{}'.format(k, delimiter, v))
        lines.append('')
        return '\n'.join(lines)


class TableFile(InputFile):

    def __init__(self, name, nrow=None):
        InputFile.__init__(self, name)
        self.table = []
        self._nrow = nrow

    @property
    def nrow(self):
        """ Return the number of rows in the table """
        if isfunction(self._nrow) | ismethod(self._nrow):
            return self._nrow(self.table)
        elif isinstance(self._nrow, int):
            return self._nrow
        else:
            return len(self.table)

    def read_string(self, value):
        self.clear()
        self.table.clear()
        lines = value.split('\n')
        for line in lines:
            if line.find(self.comment) == 0:
                pass
            elif line.find('=') > 0:
                kv = [x.strip() for x in line.split('=')]
                if len(kv) == 2:
                    self[kv[0]] = kv[1]
            else:
                try:
                    num = int(line)
                except:
                    num = None
                if num is None:
                    row = [x.strip() for x in line.split()]
                    if row:
                        self.table.append(row)

    def set_table(self, rows=None, cols=None):
        self.table.clear()
        if rows:
            self.table = rows
        elif cols:
            nrow = max([len(c) for c in cols])
            for r in range(nrow):
                row = [col[r] if r < len(col) else None for col in cols]
                self.table.append(row)

    def column(self, index, dtype='str'):
        col = [row[index] for row in self.table]
        if dtype.find('s') == 0:
            return col
        elif dtype.find('f') == 0:
            return [float(x) for x in col]
        elif dtype.find('i') == 0:
            return [int(x) for x in col]

    def to_string(self):
        lines = ['{}={}'.format(k, v) for k, v in self.items()]
        lines.append('{}'.format(self.nrow))
        for row in self.table:
            lines.append(' '.join(['{}'.format(x) for x in row]))
        lines.append('')
        return '\n'.join(lines)


class LibraryFile(InputFile):

    def __init__(self, name, key=''):
        InputFile.__init__(self, name)
        self.key = key
        self.entry = []

    def read_string(self, value):
        self.clear()
        self.entry.clear()
        lines = value.split('\n')
        entry = None
        for line in lines:
            if line.find(self.comment) == 0:
                pass
            elif line.find('=') > 0:
                kv = [x.strip() for x in line.split('=')]
                if len(kv) == 2:
                    key, val = kv
                    if key == self.key:
                        entry = OrderedDict()
                        self.entry.append(entry)
                    if entry is None:
                        self[key] = val
                    else:
                        entry[key] = val

    def to_string(self):
        lines = ['{}={}'.format(k, v) for k, v in self.items()]
        for entry in self.entry:
            lines.append(self.comment + '-----------')
            lines += ['{}={}'.format(k, v) for k, v in entry.items()]
        lines.append('')
        return '\n'.join(lines)

    def add(self, **kwargs):
        """ Add an entry in the library """
        entry = OrderedDict()
        for k, v in kwargs.items():
            entry[k.upper()] = v
        self.entry.append(entry)
        return entry


class MainFile(ConfigFile):

    def __init__(self, files, comment='!', filename='', keep_comment=False):
        ConfigFile.__init__(self, 'main', 'OUTPUT_TIMES', keep_comment)
        self.comment = comment
        for name, params in files.items():
            inp = self._input_file(name, params)
            if inp is not None:
                inp.comment = comment
                self.__dict__[name] = inp
        self.filename = filename
        if filename:
            self.read(filename)

    def _input_file(self, name, params):
        typ = ''
        arg = ''
        if isinstance(params, str):
            typ = params.lower()
        elif isinstance(params, tuple):
            typ = params[0].lower()
            arg = params[1]
        fil = None
        if typ == 'config':
            fil = ConfigFile(name, arg, self.keep_comment)
        elif typ == 'table':
            fil = TableFile(name, arg)
        elif typ == 'library':
            fil = LibraryFile(name, arg)
        return fil

    def read(self, filename):
        self.filename = filename
        ConfigFile.read(self, filename)

    def read_string(self, value):
        ConfigFile.read_string(self, value)
        dirname = os.path.dirname(self.filename)
        for key, val in self.items():
            if key.find('FILE_') == 0:
                name = key[5:].lower()
                fil = getattr(self, name, None)
                if fil is not None:
                    fil.read(os.path.join(dirname, val))

    def before_write(self):
        """ Update some inputs before writing """
        pass

    def write(self, filename):
        """ Write a input file """
        self.before_write()
        ConfigFile.write(self, filename)
        dirname = os.path.dirname(filename)
        for key, val in self.items():
            if key.find('FILE_') == 0:
                name = key[5:].lower()
                fil = getattr(self, name, None)
                if fil is not None:
                    fil.write(os.path.join(dirname, val))


class PremodInputs(MainFile):

    def __init__(self, filename=''):
        nrow = lambda table: len(table) // 2
        MainFile.__init__(self,
                          files={'solver': 'config',
                                 'alloy': 'table',
                                 'process': 'table',
                                 'phases': ('library', 'NAME'),
                                 'pptlib': ('library', 'CATEGORY'),
                                 'pptsim': ('table', nrow)},
                          filename=filename)

    def before_write(self):
        """ Update some inputs before writing """
        elements = self.alloy.column(0)
        #self.phases['NELEMENTS'] = len(elements)
        #self.phases['NPHASES'] = len(self.phases.entry)
        #self.phases['ELEMENTS'] = ', '.join([f'"{el}"' for el in elements])
        self.pptlib['NMODELS'] = len(self.pptlib.entry)

    def modify_outputtimes(self, list_times:list):
        """ modify the list of times at which microstructure state is written"""
        # sort the list
        list_times.sort()
        # convert list to string
        list_strings = [str(value) for value in list_times]

        if 'OUTPUT_TIMES' in self.keys():
            # affect the new list
            self['OUTPUT_TIMES'] = list_strings
        else:
            raise ValueError('OUTPUT_TIMES not in the keys')

    
    def modify_temperature_history(self, times_temp:list, hist_filename):
        """ modify the temperature history and possibly the name of the file"""
        # TODO need to check if the list is sorted
        # l.sort(key=lambda x: x[0]) to sort the list with the first value
        # convert list to string
        list_strings = [[str(time),str(temp)] for [time,temp] in times_temp]

        # affect the new list
        self.process.table = list_strings

        self['FILE_PROCESS'] = hist_filename


class HistoryFile(OrderedDict):

    def __init__(self, filename=''):
        self.names = []
        self.units = []
        
        if filename:
            self.read(filename)

    def get_IO_version(self,filename):
        """ get the IO_version from the header of the file"""
        self.IO_version = None
        with open(filename) as fil:
            for line in fil.readlines():
                # temporary change to handle the case
                if "_fv (-)" in line:       #"IO_version:" in line:
                    self.IO_version = 2     #int(line.strip().split(":")[1])
                    break
        if self.IO_version is None:
            self.IO_version = 1
            print("Warning: " +"IO_version not found in "+filename+ " version 1 is used")

    def clear(self):
        super().clear()
        self.names.clear()
        self.units.clear()

    def read(self, filename):
        self.get_IO_version(filename)
        if os.path.exists(filename):
            nskip = 1
            with open(filename) as fil:
                for line in fil.readlines():
                    if line[0] == "#":
                        nskip+=1
                    else:
                        names = line.split(';')
                        break
            #try:
            #print("Error in loading the file: ",filename,"\nnskip: ",nskip)
            data = np.loadtxt(filename, delimiter=';', skiprows=nskip)
            #except:
            #    print("Error in loading the file: ",filename,"\nnskip: ",nskip)
            self.clear()
            if len(data) > 0:
                for i, name in enumerate(names):
                    p1 = name.find('(')
                    p2 = name.find(')')
                    unit = ''
                    if (p1 > 0) & (p2 > p1):
                        unit = name[p1 + 1: p2].strip()
                    key = name[0:p1].strip() if p1 > 0 else name.strip()
                    self.names.append(key)
                    self.units.append(unit)
                    self[key] = data[:,i]


class Results(object):

    def __init__(self, filename='',IO_version=None):
        self.history = HistoryFile()
        self.particles = []
        self.IO_version = IO_version
        if filename:
            self.read(filename)

    def clear(self):
        self.history.clear()
        self.particles.clear()

    def _create_chem(self, ndomains, elements, phases):
        """ Create a new ChemistrySizeDistribution """
        X0 = np.zeros((len(elements), ), dtype=float)
        chem = ChemistrySizeDistribution(elements=elements,
                                         X0=X0,
                                         phases=phases)
        nelements = len(chem.elements)
        chem.Xd = np.zeros((ndomains, nelements))
        chem.Xi = np.zeros((ndomains, nelements))
        chem.rd = np.zeros((ndomains, ))
        chem.shape = np.ones((ndomains, ))
        chem.Nd = np.zeros((ndomains, ))
        width = max([len(p) for p in chem.phases])
        chem.phasenames = np.zeros((ndomains, ), dtype='U{}'.format(width))
        chem.atvol = np.ones((ndomains, ))
        chem.atvol0 = 1.0

        return chem

    def get_IO_version(self,filename):
        """ get the IO_version from the header of the file"""
        with open(filename) as fil:
            for line in fil.readlines():
                if "IO_version:" in line:
                    self.IO_version = int(line.strip().split(":")[1])
                    break
        if self.IO_version is None:
            self.IO_version = 1
            raise Warning("IO_version not found in "+filename+ " version 1 is used")

    def read_particles_v1(self, filename):
        """ Read particles file """
        with open(filename) as fil:
            lines = fil.read().split('\n')
        values = {}
        domains = []
        keys = 'ndomains,nelements,nphases,Init,Solid,Phase composition,DomainID'.split(',')
        index = {}
        for i, line in enumerate(lines):
            for key in keys:
                if line.find(key) == 0:
                    index[key] = i
            if 'DomainID' in index:
                break

        for key in ['ndomains', 'nelements', 'nphases']:
            if key in index:
                kv = lines[index[key]].split(':')
                values[key] = int(kv[1])

        i = index['nphases'] + 1
        for key in ['elements', 'phases']:
            values[key] = []
            for j in range(values['n' + key]):
                values[key].append(lines[i].strip())
                i += 1

        for key in ['Init', 'Solid', 'Phase composition']:
            values[key] = []
            i = index[key] + 1
            if key == 'Phase composition': i = index[key] + 2
            for j in range(values['nelements']):
                values[key].append(float(lines[i]))
                i += 1

        i = index['DomainID'] + 1
        for j in range(values['ndomains']):
            items = [item.strip() for item in lines[i].split()]
            domains.append([float(x) for x in items])
            i += 1
        domains = np.array(domains)

        values['phases'].insert(0, 'matrix')
        chem = self._create_chem(values['ndomains'],
                                 values['elements'],
                                 values['phases'])
        chem.X0 = np.array(values['Init'])
        for j in range(values['ndomains']):
            chem.Xd[j,:] = np.array(values['Phase composition'])
        chem.rd = domains[:,1]
        chem.shape = domains[:,2]
        chem.Nd = domains[:,3]
        phases = values['phases']
        phaseid = np.array(domains[:,4], 'i')
        for d in range(values['ndomains']):
            chem.phasenames[d] = phases[phaseid[d]]

        # deactivated as it will not reflect the content of premod state
        #chem.atvol = chem.estimated_atvol()
        #chem.atvol0 = chem.estimated_atvol('nominal')

        return chem

    def read_particles_v2(self, filename):
        """ parsing should be based on the new format"""
        metadata, df = parse_complete(filename)
        #print("*** read_particles_v2 ***\n",metadata,"\n\n",df,"\n\n")
        # post-treatment of the metadata
        # convert to float
        for key,value in metadata.items():
            if key == "X0" or "Xd_" in key or key == "XSS":
                metadata[key] = [float(val) for val in value]
            elif key in ['IO_version','ndomains','nelements','nphases']:
                metadata[key] = int(value)
        #print("*** read_particles_v2 ***\n",metadata,"\n\n",df,"\n\n")
        chem = define_chem_premod(metadata,df)

        return chem

    def read_particles(self,filename):
        """ read the particles data depending on the IO_version"""
        self.get_IO_version(filename)

        #print("*********** filename=",filename)
        #print("IO_version=",self.IO_version)
        if self.IO_version == 1:
            return self.read_particles_v1(filename)
        elif self.IO_version == 2:
            return self.read_particles_v2(filename)
        else:
            raise ValueError("In read_particles, IO_version number not valid: "
                            ,self.IO_version)


    def read_particle_evolution(self, pattern, times, temperatures):
        """ Read particles outputs """
        self.particles.clear()
        for filename in glob(pattern):
            basename = os.path.splitext(os.path.basename(filename))[0]
            parts = basename.split('Micro_')
            #raise Exception(basename, pattern, parts)
            if len(parts) == 2:
                chem = self.read_particles(filename)
                if chem:
                    tim = float(parts[1])
                    item = {'time': tim,
                            'temperature': np.interp(tim, times, temperatures),
                            'chem': chem}
                    self.particles.append(item)
        self.particles.sort(key=lambda item: item['time'])

    def read(self, filename, inp=None):
        dirname = os.path.dirname(filename)
        base = os.path.splitext(os.path.basename(filename))[0]
        ppt = os.path.join(dirname, base, base + '_PPT.txt')
        self.history.read(ppt)
        micro = os.path.join(dirname, base, base + '_Micro_*.txt')
        if inp is None:
            inp = PremodInputs(filename)
        times = np.array(inp.process.column(0, 'float'))
        temperatures = np.array(inp.process.column(1, 'float'))
        self.read_particle_evolution(micro, times, temperatures)



#!/usr/bin/env python

from setuptools import setup

setup(name='pypremod',
      description='Python package for Premod solver',
      url='https://www.sintef.no',
      install_requires=[
          'numpy',
          'pandas',
          'pytest',
          'scipy',
          'matplotlib',
          'pypremod-calm',
          'pypremod-strength'
      ],
      version = '0.1.3',
      packages=['premod'])
